@@ -7,12 +7,12 @@ demo applications based on the gui_app and the related GUI framework implementat
 (like e.g. ae.kivy_app and ae.enaml_app) of the ae namespace.
 """
 from abc import abstractmethod, ABC
-from typing import Any, Dict, List, Optional
+from typing import Any, Dict, List, Optional, Sequence
 
-from ae.gui_app import id_of_flow, flow_action, flow_key        # type: ignore
+from ae.gui_app import id_of_flow, flow_action, flow_key, replace_flow_action        # type: ignore
 
 
-__version__ = '0.0.1'
+__version__ = '0.0.2'
 
 
 FLOW_PATH_ROOT_ID = '<<<root>>>'     #: needed for flow path jumper
@@ -35,7 +35,7 @@ HELP_TEXT = (
 )
 
 
-LiszItem = Dict[str, Any]               #: list item data (lid) type
+LiszItem = Dict[str, Any]               #: item data (lid) type
 LiszNode = List[LiszItem]               #: node/list type
 
 
@@ -51,25 +51,116 @@ def check_item_id(item_id: str) -> bool:
 
 class LiszDataMixin(ABC):
     """ lisz data model - independent from used GUI framework. """
-    root_node: LiszNode             #: root of complete/full list data structure
+    root_node: LiszNode             #: root of lisz data structure
     current_node: LiszNode          #: currently selected/displayed node / sub list
 
-    # mixin shadow attributes - implemented by :class:`~ae.gui_app.MainAppBase`
+    # mixin shadow attributes - implemented by :class:`~ae.console_app.ConsoleApp` or :class:`~ae.gui_app.MainAppBase`
+    debug_level: int                #: :attr:`~AppBase.debug_level`
     flow_id: str                    #: current attr:`flow id <ae.gui_app.MainAppBase.flow_id>`
     flow_path: List[str]            #: :attr:`flow path <ae.gui_app.MainAppBase.flow_path>` ref. current node
-    debug_level: int                #: :attr:`~AppBase.debug_level`
+    framework_win: Any              #: framework window
+
+    _refreshing_data: bool = False                      #: DEBUG True while running :meth:`~.refresh_all` method
+
+    @abstractmethod
+    def call_method(self, method: str, *args, **kwargs) -> Any:
+        """ call method of this instance (ae.gui_app.MainAppBase method). """
 
     @abstractmethod
     def change_app_state(self, state_name: str, new_value: Any, send_event: bool = True):
-        """ add debug sound on each state change/save. """
+        """ add debug sound on each state change/save (ae.gui_app.MainAppBase method). """
+
+    @abstractmethod
+    def change_flow(self, ifi: str, popups_to_close: Sequence = (), **event_kwargs) -> bool:
+        """ change/switch flow id. """
 
     @abstractmethod
     def flow_path_action(self, path_index: int = -1) -> str:
-        """ determine the action of the last/newest entry in the flow_path. """
+        """ determine the action of the last/newest entry in the flow_path (ae.gui_app.MainAppBase method). """
 
     @abstractmethod
     def play_sound(self, sound_name: str):
-        """ play audio/sound file. """
+        """ play audio/sound file (ae.gui_app.MainAppBase method). """
+
+    @abstractmethod
+    def refresh_node_widgets(self):
+        """ redraw/refresh the widgets representing the current node items/sub-nodes (GUI framework method). """
+
+    # helper methods
+
+    def delete_item(self, item_id: str, node_only: bool):
+        """ delete either the complete item or the sub node of the item (identified by the passed item id).
+
+        :param item_id:         item id to identify the item/sub-node to be deleted.
+        :param node_only:       True if only delete the sub-node of the identified item, False for to delete the item.
+        """
+        lid = self.item_by_id(item_id)
+        if node_only:
+            lid.pop('node')
+        else:
+            self.current_node.remove(lid)
+            self.change_app_state('flow_id', id_of_flow('', ''), send_event=False)
+
+        self.play_sound('deleted')
+
+    def edit_validate(self, old_idx: int, new_id: str, want_node: bool) -> str:
+        """ validate the user changes after adding a new item or editing an existing item.
+
+        :param old_idx:         index in the current node of the edited item or -1 if a new item got edited.
+        :param new_id:          new/edited id string.
+        :param want_node:       True if the new/edited item will have a sub-node.
+        :return:                empty string on successful validation, else error string or
+                                `'request_delete_confirmation_for_item'` if the user has to confirm the deletion
+                                after the user wiped the item id string or
+                                `'request_delete_confirmation_for_node'` if the user has to confirm the
+                                removal of the sub-node.
+        """
+        if not check_item_id(new_id):
+            return f"invalid item name/id {new_id} containing '{FLOW_PATH_TEXT_SEP}' or is equal '{FLOW_PATH_ROOT_ID}'"
+
+        add_item = old_idx == -1
+        if not new_id:
+            # on empty id cancel edit if add_item, else request confirmation from user for item del
+            return "" if add_item else 'request_delete_confirmation_for_item'
+
+        new_idx = self.find_item_index(new_id)
+        if new_idx != -1 and (add_item or new_idx != old_idx):
+            return f"item id {new_id} exists already"
+
+        if add_item:                        # add new item
+            lid = dict(id=new_id)
+            if want_node:
+                lid['node'] = list()        # type: ignore   # mypy not supports recursive types see issue #731
+            self.current_node.insert(0, lid)
+
+        else:                               # edit item
+            lid = self.current_node[old_idx]
+            lid['id'] = new_id
+            if want_node != ('node' in lid):    # NOTE: != operator has lower priority than in operator
+                if want_node:
+                    lid['node'] = list()        # type: ignore   # mypy not supports recursive types see issue #731
+                elif lid['node']:       # let user confirm node deletion of non-empty lid['node']
+                    return 'request_delete_confirmation_for_node'
+                else:
+                    lid.pop('node')     # remove empty node
+
+        self.play_sound('added' if add_item else 'edited')
+
+        return ""
+
+    def find_item_index(self, item_id: str, searched_node: Optional[LiszNode] = None) -> int:
+        """ determine list index of the passed item id in the searched node or in the current node.
+
+        :param item_id:         item id to find.
+        :param searched_node:   searched node. if not passed then the current node will be searched instead.
+        :return:                item list index in the searched node or -1 if item id was not found.
+        """
+        if searched_node is None:
+            searched_node = self.current_node
+        for list_idx, lid in enumerate(searched_node):
+            if lid['id'] == item_id:
+                return list_idx
+        return -1
 
     def flow_key_text(self, flow_id: str, landscape: bool) -> str:
         """ determine shortest possible text fragment of the passed flow key that is unique in the current node.
@@ -94,28 +185,6 @@ class LiszDataMixin(ABC):
             return f" ->{key[:id_len]}"
         return f".{flow_id}" if self.debug_level else ""
 
-    def flow_path_lists(self):
-        """ determine nodes relative to the current flow path to quick-jump to from current list. """
-        def append_nodes(node):
-            """ recursively collect all available nodes (possible flow paths) """
-            for lid in node:
-                if 'node' in lid:
-                    deeper_flow_path.append(id_of_flow('enter', 'item', lid['id']))
-                    paths.append(self.flow_path_text(deeper_flow_path))
-                    append_nodes(lid['node'])
-                    deeper_flow_path.pop()
-
-        paths = [FLOW_PATH_ROOT_ID] if self.flow_path else list()
-        deep = 0
-        while deep < len(self.flow_path) and self.flow_path_action(deep) == 'enter':
-            deep += 1
-            paths.append(self.flow_path_text(self.flow_path[:deep]))
-
-        deeper_flow_path = self.flow_path[:deep]
-        append_nodes(self.current_node)
-
-        return paths
-
     def flow_path_from_text(self, text: str) -> List[str]:
         """ restore the full complete flow path from the shortened flow keys generated by :meth:`.flow_path_text`.
 
@@ -178,106 +247,43 @@ class LiszDataMixin(ABC):
             if flow_action(flow_id) != 'enter':
                 break
             node_id = flow_key(flow_id)
-            current_node = self.item_by_id(node_id, searched_list=current_node)['node']
+            current_node = self.item_by_id(node_id, searched_node=current_node)['node']
         return current_node
 
-    def delete_item(self, item_id: str, node_only: bool):
-        """ delete either the complete item or the sub node of the item (identified by the passed item id).
+    def flow_path_quick_jump_nodes(self):
+        """ determine nodes relative to the current flow path to quick-jump to from current node. """
+        def append_nodes(node):
+            """ recursively collect all available nodes (possible flow paths) """
+            for lid in node:
+                if 'node' in lid:
+                    deeper_flow_path.append(id_of_flow('enter', 'item', lid['id']))
+                    paths.append(self.flow_path_text(deeper_flow_path))
+                    append_nodes(lid['node'])
+                    deeper_flow_path.pop()
 
-        :param item_id:         item id to identify the item/sub-node to be deleted.
-        :param node_only:       True if only delete the sub-node of the identified item, False for to delete the item.
-        """
-        lid = self.item_by_id(item_id)
-        if node_only:
-            lid.pop('node')
-        else:
-            self.current_node.remove(lid)
-            self.change_app_state('flow_id', id_of_flow('', ''), send_event=False)
+        paths = [FLOW_PATH_ROOT_ID] if self.flow_path else list()
+        deep = 0
+        while deep < len(self.flow_path) and self.flow_path_action(deep) == 'enter':
+            deep += 1
+            paths.append(self.flow_path_text(self.flow_path[:deep]))
 
-        self.play_sound('deleted')
+        deeper_flow_path = self.flow_path[:deep]
+        append_nodes(self.current_node)
 
-    def edit_validate(self, old_idx: int, new_id: str, want_list: bool) -> str:
-        """ validate the user changes after adding a new item or editing an existing item.
+        return paths
 
-        :param old_idx:         index in the current node of the edited item or -1 if a new item got edited.
-        :param new_id:          new/edited id string.
-        :param want_list:       True if the new/edited item will have a sub-node.
-        :return:                empty string on successful validation, else error string or
-                                `'request_delete_confirmation_for_item'` if the user has to confirm the deletion
-                                after the user wiped the item id string or
-                                `'request_delete_confirmation_for_node'` if the user has to confirm the
-                                removal of the sub-node.
-        """
-        if not check_item_id(new_id):
-            return f"invalid item name/id {new_id} containing '{FLOW_PATH_TEXT_SEP}' or is equal '{FLOW_PATH_ROOT_ID}'"
-
-        add_item = old_idx == -1
-        if not new_id:
-            # on empty id cancel edit if add_item, else request confirmation from user for item del
-            return "" if add_item else 'request_delete_confirmation_for_item'
-
-        new_idx = self.find_item_index(new_id)
-        if new_idx != -1 and (add_item or new_idx != old_idx):
-            return f"item id {new_id} exists already"
-
-        if add_item:                        # add new item
-            lid = dict(id=new_id)
-            if want_list:
-                lid['node'] = list()        # type: ignore   # mypy not supports recursive types see issue #731
-            self.current_node.insert(0, lid)
-
-        else:                               # edit item
-            lid = self.current_node[old_idx]
-            lid['id'] = new_id
-            if want_list != ('node' in lid):    # NOTE: != operator has lower priority than in operator
-                if want_list:
-                    lid['node'] = list()        # type: ignore   # mypy not supports recursive types see issue #731
-                elif lid['node']:       # let user confirm list deletion of non-empty lid['node']
-                    return 'request_delete_confirmation_for_node'
-                else:
-                    lid.pop('node')     # remove empty node
-
-        self.play_sound('added' if add_item else 'edited')
-
-        return ""
-
-    def filter_toggle(self, toggle_attr: str, invert_attr: str):
-        """ toggle the item filter and ensure that the inverted filter will be toggled to prevent both filters active.
-
-        :param toggle_attr:     filter for to toggle.
-        :param invert_attr:     inverted filter will be set to False if was True and toggled filter get changed to True.
-        """
-        filtering = not getattr(self, toggle_attr)
-        self.change_app_state(toggle_attr, filtering)
-        if filtering and getattr(self, invert_attr):
-            self.change_app_state(invert_attr, False)
-
-    def find_item_index(self, item_id: str, searched_list: Optional[LiszNode] = None) -> int:
-        """ determine list index of the passed item id in the searched list or in the current node.
+    def item_by_id(self, item_id: str, searched_node: Optional[LiszNode] = None) -> LiszItem:
+        """ search item in either the passed or the current node.
 
         :param item_id:         item id to find.
-        :param searched_list:   searched node. if not passed then the current node will be searched instead.
-        :return:                item list index in the searched list or -1 if item id was not found.
-        """
-        if searched_list is None:
-            searched_list = self.current_node
-        for list_idx, lid in enumerate(searched_list):
-            if lid['id'] == item_id:
-                return list_idx
-        return -1
-
-    def item_by_id(self, item_id: str, searched_list: Optional[LiszNode] = None) -> LiszItem:
-        """ search list item in either the passed or the current node.
-
-        :param item_id:         item id to find.
-        :param searched_list:   searched node. if not passed then the current node will be searched instead.
+        :param searched_node:   searched node. if not passed then the current node will be searched instead.
         :return:                found item or empty dict with new/empty id if not found.
         """
-        if searched_list is None:
-            searched_list = self.current_node
-        index = self.find_item_index(item_id, searched_list=searched_list)
+        if searched_node is None:
+            searched_node = self.current_node
+        index = self.find_item_index(item_id, searched_node=searched_node)
         if index != -1:
-            return searched_list[index]
+            return searched_node[index]
         return dict(id='')
 
     def move_item(self, dragged_node: LiszNode, item_id: str, node_idx: int):
@@ -287,7 +293,7 @@ class LiszDataMixin(ABC):
         :param item_id:         id of the moved item.
         :param node_idx:        index in the node where the item has to be moved to.
         """
-        src_idx = self.find_item_index(item_id, searched_list=dragged_node)
+        src_idx = self.find_item_index(item_id, searched_node=dragged_node)
         lid = dragged_node.pop(src_idx)
         if dragged_node == self.current_node and node_idx > src_idx:
             node_idx -= 1
@@ -295,7 +301,99 @@ class LiszDataMixin(ABC):
 
     def on_app_init(self):
         """ initialize self.current_node from flow_id/flow_path saved on last app stop. """
-        self.refresh_current_node()
+        self.refresh_current_node_from_flow_path()
+
+    def on_filter_toggle(self, flow_id: str, _event_kwargs: Dict[str, Any]) -> bool:
+        """ toggle filter on click of either the selected or the unselected filter button.
+
+        Note that the inverted filter may be toggled to prevent both filters active.
+
+        :param flow_id:         flow id, flow key is specifying the filter button.
+        :param _event_kwargs:   unused.
+        :return:                True to process flow id change.
+        """
+        toggle_attr = flow_key(flow_id)     # filter to toggle (either 'filter_selected' or 'filter_unselected')
+        # inverted filter will be set to False if was True and toggled filter get changed to True.
+        invert_attr = 'filter_unselected' if toggle_attr == 'filter_selected' else 'filter_selected'
+
+        filtering = not getattr(self, toggle_attr)
+        self.change_app_state(toggle_attr, filtering)
+        if filtering and getattr(self, invert_attr):
+            self.change_app_state(invert_attr, False)
+
+        self.play_sound(f'filter_{"on" if filtering else "off"}')
+        self.refresh_node_widgets()
+
+        return True
+
+    def on_key_press(self, modifiers: str, key_code: str) -> bool:
+        """  check key press event for to be handled and processed as command/action.
+
+        :param modifiers:   modifier keys string.
+        :param key_code:    key code string.
+        :return:            True if key event got processed/used, else False.
+        """
+        if modifiers or flow_action(self.flow_id) not in ('', 'focus'):
+            return False
+
+        # handle hot key without a modifier key and while in item list, first current item flow changes
+        flo_key = flow_key(self.flow_id)
+        if key_code == 'up':
+            self.set_neighbour_flow(-1)
+        elif key_code == 'down':
+            self.set_neighbour_flow(1)
+        elif key_code == 'pgup':
+            self.set_neighbour_flow(-15)
+        elif key_code == 'pgdown':
+            self.set_neighbour_flow(15)
+        elif key_code == 'home':
+            self.set_neighbour_flow(-999999)
+        elif key_code == 'end':
+            self.set_neighbour_flow(999999)
+
+        # toggle selection of current item
+        elif key_code == ' ' and self.flow_id:    # key string 'space' is not in Window.command_keys
+            self.toggle_item_sel(self.find_item_index(flo_key))
+            self.refresh_node_widgets()
+
+        # enter/leave flow in current list
+        elif key_code in ('enter', 'right') and self.flow_id and 'node' in self.item_by_id(flo_key):
+            self.change_flow(id_of_flow('enter', 'node', flo_key))
+        elif key_code in ('escape', 'left') and self.flow_path:
+            self.change_flow(id_of_flow('leave', 'node'))
+
+        # item processing: add, edit or request confirmation of deletion of current item
+        elif key_code in ('a', '+'):
+            self.change_flow(id_of_flow('add', 'item'))
+
+        elif key_code == 'e' and self.flow_id:
+            self.change_flow(replace_flow_action(self.flow_id, 'edit'))  # popup_kwargs=dict(parent=self.framework_win))
+
+        elif key_code in ('-', 'del') and self.flow_id:
+            self.change_flow(id_of_flow('confirm', 'item_deletion'), item_id=flo_key)
+
+        else:
+            return False    # pressed key not processable in the current flow/app-state
+
+        return True         # key press processed
+
+    def on_node_enter(self, _flow_id: str, _event_kwargs: dict):
+        """ entering sub node from current node.
+
+        :param _flow_id:        flow id.
+        :param _event_kwargs:   event kwargs.
+        :return:                True for to process/change flow id.
+        """
+        self.play_sound(id_of_flow('enter', 'node'))
+
+    def on_node_leave(self, _flow_id: str, _event_kwargs: dict):
+        """ leaving sub node, setting current node to parent node.
+
+        :param _flow_id:        flow id.
+        :param _event_kwargs:   event kwargs.
+        :return:                True for to process/change flow id.
+        """
+        self.play_sound(id_of_flow('leave', 'node'))
 
     def on_node_jump(self, flow_id: str, _event_kwargs: dict) -> bool:
         """ FlowButton clicked event handler restoring flow path from the flow key.
@@ -318,12 +416,52 @@ class LiszDataMixin(ABC):
         self.change_app_state('flow_path', flow_path)
         self.change_app_state('flow_id', id_of_flow('', ''))
 
+        self.play_sound(id_of_flow('enter', 'node'))
+        self.refresh_all()
         return True
 
-    def refresh_current_node(self):
+    def refresh_all(self):
+        """ refresh currently displayed items after changing current node. """
+        assert not self._refreshing_data
+        self._refreshing_data = True
+        if self.debug_level:
+            self.play_sound('debug_draw')
+
+        self.refresh_current_node_from_flow_path()
+
+        # save last actual flow id (because refreshed/redrawn widget observers could change flow id via focus)
+        flow_id = self.flow_id
+
+        self.refresh_node_widgets()
+
+        redraw = False
+        if flow_action(flow_id) == 'focus' and self.find_item_index(flow_key(flow_id)) == -1:
+            flow_id = id_of_flow('', '')    # reset flow id because last focused item got filtered/deleted by user
+            redraw = True
+
+        self.change_app_state('flow_id', flow_id, send_event=redraw)    # correct flow or restore silently
+
+        if flow_action(flow_id) == 'focus':
+            self.call_method('on_flow_widget_focused')                  # restore focus
+
+        assert self._refreshing_data
+        self._refreshing_data = False
+
+    def refresh_current_node_from_flow_path(self):
         """ refresh current node including the depending display node. """
         self.current_node = self.flow_path_node()
 
+    def set_neighbour_flow(self, delta):
+        """ move flow id to previous/next item. """
+        current_node = self.current_node
+        if current_node:
+            flow_id = self.flow_id
+            if flow_id:
+                idx = min(max(0, self.find_item_index(flow_key(flow_id)) + delta), len(current_node) - 1)
+            else:
+                idx = min(max(-1, delta), 0)
+            self.change_app_state('flow_id', id_of_flow('focus', 'item', current_node[idx]['id']))
+
     def sub_item_ids(self, item_id, node_only, node=None, sub_item_ids=None):
         """ return item names of all items including their sub-node items (if exists).
 
@@ -353,12 +491,12 @@ class LiszDataMixin(ABC):
 
         return sub_item_ids
 
-    def update_item_sel(self, new_sel: int, node_idx: int):
+    def toggle_item_sel(self, node_idx: int):
         """ update the item selection data of the item identified by the list index in the current node.
 
-        :param new_sel:         new selection: True if selected, False if not.
         :param node_idx:        list index of the item in the current node to change the selection for.
         """
+        new_sel = not self.current_node[node_idx].get('sel', False)
         if new_sel:
             self.current_node[node_idx]['sel'] = 1
         else:
